#!/usr/bin/env python3
"""
## Network and Disk I/O Stats Monitor

Monitor and log Network and Disks statistics in MegaBytes per second.

### Pytorch-lightning integration

```py
# TODO: add a pytorch-lightning integration example
```

### Pure-Python implementation

```py
import time
from iometrics import NetworkMetrics, DiskMetrics, DUAL_METRICS_HEADER
net  = NetworkMetrics()
disk = DiskMetrics()
for i in range(100):
    time.sleep(1)
    net.update_stats()
    disk.update_stats()
    print(DUAL_METRICS_HEADER) if i % 15 == 0 else None
    row = (
        f"| {net.mb_recv_ps.val:6.1f} | {net.mb_recv_ps.avg:6.1f} "
        f"| {net.mb_sent_ps.val:5.1f} | {net.mb_sent_ps.avg:5.1f} "
        f"| {int(disk.io_util.val):3d} | {int(disk.io_util.avg):3d} "
        f"| {disk.mb_read.val:6.1f} | {disk.mb_read.avg:6.1f} "
        f"| {disk.mb_writ.val:5.1f} | {disk.mb_writ.avg:5.1f} "
        f"| {int(disk.io_read.val):4d} | {int(disk.io_read.avg):4d} "
        f"| {int(disk.io_writ.val):3d} | {int(disk.io_writ.avg):3d} "
        f"|"
    )
    print(row)
```

#### Example output

```md
|        Network (MBytes/s)       | Disk Util |            Disk MBytes          |           Disk I/O          |
|     Received    |     Sent      |     %     |    MB/s Read    |  MB/s Written |     I/O Read    | I/O Write |
|   val  |   avg  |  val  |  avg  | val | avg |  val   |  avg   |  val  |  avg  |   val  |   avg  | val | avg |
| ------:| ------:| -----:| -----:| ---:| ---:| ------:| ------:| -----:| -----:| ------:| ------:| ---:| ---:|
|    4.6 |    3.5 |   0.1 |   0.1 |  49 |   2 |   52.8 |    1.1 |   0.0 |   0.9 |    211 |      4 |   5 |  18 |
|    4.1 |    3.5 |   0.1 |   0.1 |  61 |   3 |   60.4 |    2.4 |  40.3 |   1.7 |    255 |     10 | 149 |  21 |
```

:copyright: (c) 2021 by Leo Gallucci.
:license: Apache 2.0, see LICENSE for more details.
"""
# Do not change the version here but rather `tbump "0.0.5" --only-patch`
__version__ = "0.0.1"

from iometrics.network import NetworkMetrics
from iometrics.disk import DiskMetrics
from iometrics.example import usage, DUAL_METRICS_HEADER

# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "__version__",
    "usage",
    "DUAL_METRICS_HEADER",
    "NetworkMetrics",
    "DiskMetrics",
]



#!/usr/bin/env python3
import re
import time
from typing import Dict
from dataclasses import dataclass

from iometrics.common import AverageMetrics


@dataclass
class NetworkStats:
    """Stores network I/O statistics."""
    bytes_recv: int = 0
    bytes_sent: int = 0

class NetworkMetrics:
    def __init__(self) -> None:
        self.mb_recv_ps = AverageMetrics()
        self.mb_sent_ps = AverageMetrics()

        self.last_stats: Dict[str, NetworkStats] = self.get_network_bytes()
        self.last_log_time: float = time.time()

    def get_network_bytes(self) -> Dict[str, NetworkStats]:
        """Returns received, transmitted bytes (since the kernel started)."""
        # Important: You should wait at least 1 second between calls to this function.
        # Before ~0.9 seconds `/proc/net/dev` will show the exact same values as last the time.

        INTERFACE_FILTER_OUT = re.compile(r'^(lo|tun.+|face.+|bond.+|.+\.\d+)$')

        # Note: all counters at /proc/* are starting with zero when the kernel starts.
        with open('/proc/net/dev') as f1:
            content: str = f1.read()

        lines: list = content.splitlines()
        lines = lines[2:]  # strip header

        # Accumulate the stats of each device:
        stats: Dict[str, NetworkStats] = {}

        for device_line in lines:
            fields: list = device_line.strip().split()
            device_name: str = fields[0].strip(':')

            if INTERFACE_FILTER_OUT.match(device_name):
                continue

            device_stats = NetworkStats(
                # Using max because sometimes it returns a negative (wrong) number
                bytes_recv = max(0, int(fields[1])),
                bytes_sent = max(0, int(fields[9])),
            )

            stats[device_name] = device_stats

        return stats

    def update_stats(self) -> None:
        """Computes metrics since last measurement then returns stats per second."""
        time_delta: float = time.time() - self.last_log_time

        # Important: You should wait at least 1 second between calls to the get_network_bytes function.
        # Before ~0.9 seconds `/proc/net/dev` will show the exact same values as last the time.
        if time_delta < 0.99:
            return None

        new_stats: Dict[str, NetworkStats] = self.get_network_bytes()

        aggr_mb_recv_ps: float = 0.0
        aggr_mb_sent_ps: float = 0.0

        # Note: iterate over last_stats instead of new_stats becase a new device
        #       could have been plugged and there won't be previous stats (baseline)
        # TODO: However, this might raise a KeyError if a device gets unplugged.
        for device_name, last_stats in self.last_stats.items():
            bytes_recv_delta: int = new_stats[device_name].bytes_recv - last_stats.bytes_recv
            bytes_sent_delta: int = new_stats[device_name].bytes_sent - last_stats.bytes_sent

            # There's a bug that sometimes the delta is negative messing up the average.
            bytes_recv_delta = max(0, bytes_recv_delta)
            bytes_sent_delta = max(0, bytes_sent_delta)

            mb_recv_ps: float = bytes_recv_delta / 1e6 / time_delta
            mb_sent_ps: float = bytes_sent_delta / 1e6 / time_delta

            aggr_mb_recv_ps += mb_recv_ps
            aggr_mb_sent_ps += mb_sent_ps

        self.mb_recv_ps.update(aggr_mb_recv_ps)
        self.mb_sent_ps.update(aggr_mb_sent_ps)

        self.last_log_time = time.time()
        self.last_stats = new_stats

# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "NetworkMetrics",
]



#!/usr/bin/env python3
from dataclasses import dataclass


@dataclass
class AverageMetrics:
    """Computes and stores the average and current value of some metric."""
    avg_mom:    float = 0.5
    val:        float = 0.0
    avg:        float = 0.0
    smooth_avg: float = 0.0
    tot_sum:    float = 0.0
    count:      int = 0

    def __init__(self, avg_mom: float = 0.5) -> None:
        self.avg_mom = avg_mom
        self.reset()

    def reset(self) -> None:
        """Reset values when starting a new epoch (for example)."""
        self.val = 0.0
        self.avg = 0.0
        self.smooth_avg = 0.0
        self.tot_sum = 0.0
        self.count = 0

    def update(self, val: float) -> None:
        # Use max(0) to prevent rounding -0.0 negative numbers
        self.val = max(0, val)
        self.tot_sum += val
        if self.count == 0:
            self.smooth_avg = val
        else:
            self.smooth_avg = self.avg * self.avg_mom + val * (1 - self.avg_mom)
        self.count += 1
        self.avg = self.tot_sum / self.count


# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "AverageMetrics",
]



#!/usr/bin/env python3
import os
import re
import time
from typing import List, Dict
from dataclasses import dataclass

from iometrics.common import AverageMetrics


@dataclass
class DiskStats:
    """Stores Disk I/O statistics."""
    kb_read: int = 0
    kb_writ: int = 0
    io_read: int = 0
    io_writ: int = 0
    io_util: int = 0

class DiskMetrics:
    def __init__(self) -> None:
        self.mb_read = AverageMetrics()
        self.mb_writ = AverageMetrics()
        self.io_read = AverageMetrics()
        self.io_writ = AverageMetrics()
        self.io_util = AverageMetrics()

        self.non_virtual_devices = self.get_non_virtual_disk_devices()

        self.last_stats: Dict[str, DiskStats] = self.get_disks_stats()
        self.last_log_time: float = time.time()

    def get_non_virtual_disk_devices(self) -> List[str]:
        """Returns a list of currently attached Disk devices names to which it makes sense to measure I/O performance."""
        DISKS_DEVICES_DIR = "/sys/block"

        non_virtual_devices: List[str] = []

        for device_name in os.listdir(DISKS_DEVICES_DIR):
            abs_device_path = os.path.join(DISKS_DEVICES_DIR, device_name)

            if os.path.islink(abs_device_path):
                link_target = os.readlink(abs_device_path)
                if 'virtual' not in link_target:
                    non_virtual_devices.append(device_name)

        return non_virtual_devices

    def get_disks_stats(self) -> Dict[str, DiskStats]:
        """Returns number of disk reads, writes, io (since the kernel started)."""

        # Note: all counters at /proc/* are starting with zero when the kernel starts.
        with open('/proc/diskstats') as f1:
            content: str = f1.read()

        lines = content.splitlines()

        # Accumulate the stats of each device:
        stats: Dict[str, DiskStats] = {}

        for device_line in lines:
            fields: list = device_line.strip().split()
            device_name: str = fields[2]

            if device_name not in self.non_virtual_devices:
                continue

            device_stats = DiskStats(
                io_read = int(fields[3]),
                io_writ = int(fields[7]),
                io_util = int(fields[12]),
                kb_read = int(fields[5]),
                kb_writ = int(fields[9]),
            )

            stats[device_name] = device_stats

        return stats

    def update_stats(self) -> None:
        """Computes metrics since last measurement then returns stats per second."""
        time_delta: float = time.time() - self.last_log_time

        new_stats: Dict[str, DiskStats] = self.get_disks_stats()

        aggr_mb_read_ps: float = 0.0
        aggr_mb_writ_ps: float = 0.0
        aggr_io_read_ps: float = 0.0
        aggr_io_writ_ps: float = 0.0
        aggr_io_util:    float = 0.0

        for device_name, last_stats in self.last_stats.items():
            bytes_read_delta: int = new_stats[device_name].kb_read - last_stats.kb_read
            bytes_writ_delta: int = new_stats[device_name].kb_writ - last_stats.kb_writ

            # There's a bug that sometimes the delta is negative messing up the average.
            bytes_read_delta = max(0, bytes_read_delta)
            bytes_read_delta = max(0, bytes_read_delta)

            mb_read_ps: float = bytes_read_delta * 512.0 / 1e6 / time_delta
            mb_writ_ps: float = bytes_writ_delta * 512.0 / 1e6 / time_delta

            io_read_delta: int = new_stats[device_name].io_read - last_stats.io_read
            io_writ_delta: int = new_stats[device_name].io_writ - last_stats.io_writ

            # There's a bug that sometimes the delta is negative messing up the average.
            io_read_delta = max(0, io_read_delta)
            io_writ_delta = max(0, io_writ_delta)

            io_read_ps = io_read_delta / time_delta
            io_writ_ps = io_writ_delta / time_delta

            util_delta: int = new_stats[device_name].io_util - last_stats.io_util
            util_delta = max(0, util_delta)
            util = 100 * util_delta / (time_delta * 1000.0)

            aggr_mb_read_ps += mb_read_ps
            aggr_mb_writ_ps += mb_writ_ps
            aggr_io_read_ps += io_read_ps
            aggr_io_writ_ps += io_writ_ps
            aggr_io_util    += min(100, util)

        avg_disks_io_util: float = aggr_io_util / len(self.last_stats)

        self.mb_read.update(aggr_mb_read_ps)
        self.mb_writ.update(aggr_mb_writ_ps)
        self.io_read.update(aggr_io_read_ps)
        self.io_writ.update(aggr_io_writ_ps)
        self.io_util.update(avg_disks_io_util)

        self.last_log_time = time.time()
        self.last_stats = new_stats

# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "DiskMetrics",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['iometrics']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'iometrics',
    'version': '0.0.1',
    'description': 'Network and Disk I/O Stats Monitor',
    'long_description': '# iometrics\n\n[![Python](docs/img/badges/language.svg)](https://devdocs.io/python/)\n\nNetwork and Disk I/O Stats Monitor.\n\n## Usage\n\nMonitor and log Network and Disks statistics in MegaBytes per second.\n\n### Pytorch-lightning integration\n\n```py\n# TODO: add a pytorch-lightning integration example\n```\n\n### Pure-Python implementation\n\nQuick check:\n\n```sh\npython -c \'import iometrics; iometrics.usage()\'\n```\n\nFull code:\n\n```py\nimport time\nfrom iometrics import NetworkMetrics, DiskMetrics, DUAL_METRICS_HEADER\nnet  = NetworkMetrics()\ndisk = DiskMetrics()\nfor i in range(100):\n    time.sleep(1)\n    net.update_stats()\n    disk.update_stats()\n    print(DUAL_METRICS_HEADER) if i % 15 == 0 else None\n    row = (\n        f"| {net.mb_recv_ps.val:6.1f} | {net.mb_recv_ps.avg:6.1f} "\n        f"| {net.mb_sent_ps.val:5.1f} | {net.mb_sent_ps.avg:5.1f} "\n        f"| {int(disk.io_util.val):3d} | {int(disk.io_util.avg):3d} "\n        f"| {disk.mb_read.val:6.1f} | {disk.mb_read.avg:6.1f} "\n        f"| {disk.mb_writ.val:5.1f} | {disk.mb_writ.avg:5.1f} "\n        f"| {int(disk.io_read.val):4d} | {int(disk.io_read.avg):4d} "\n        f"| {int(disk.io_writ.val):3d} | {int(disk.io_writ.avg):3d} "\n        f"|"\n    )\n    print(row)\n```\n\n#### Example output\n\n```markdown\n|        Network (MBytes/s)       | Disk Util |            Disk MBytes          |           Disk I/O          |\n|     Received    |     Sent      |     %     |    MB/s Read    |  MB/s Written |     I/O Read    | I/O Write |\n|   val  |   avg  |  val  |  avg  | val | avg |  val   |  avg   |  val  |  avg  |   val  |   avg  | val | avg |\n| ------:| ------:| -----:| -----:| ---:| ---:| ------:| ------:| -----:| -----:| ------:| ------:| ---:| ---:|\n|    4.6 |    3.5 |   0.1 |   0.1 |  49 |   2 |   52.8 |    1.1 |   0.0 |   0.9 |    211 |      4 |   5 |  18 |\n|    4.1 |    3.5 |   0.1 |   0.1 |  61 |   3 |   60.4 |    2.4 |  40.3 |   1.7 |    255 |     10 | 149 |  21 |\n```\n\n## Contributing\n\nSee [CONTRIBUTING.md](CONTRIBUTING.md)\n',
    'author': 'Leo Gallucci',
    'author_email': 'elgalu3@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/elgalu/iometrics',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
import time

from iometrics import NetworkMetrics, DiskMetrics


DUAL_METRICS_HEADER = """
|        Network (MBytes/s)       | Disk Util |            Disk MBytes          |             Disk I/O            |
|     Received    |     Sent      |     %     |    MB/s Read    |  MB/s Written |     I/O Read    |   I/O Write   |
|   val  |   avg  |  val  |  avg  | val | avg |  val   |  avg   |  val  |  avg  |   val  |   avg  |  val  |  avg  |
| ------:| ------:| -----:| -----:| ---:| ---:| ------:| ------:| -----:| -----:| ------:| ------:| -----:| -----:|"""


def usage() -> None:
    """
    ## Example usage and example output:

    ```markdown
    |        Network (MBytes/s)       | Disk Util |            Disk MBytes          |           Disk I/O          |
    |     Received    |     Sent      |     %     |    MB/s Read    |  MB/s Written |     I/O Read    | I/O Write |
    |   val  |   avg  |  val  |  avg  | val | avg |  val   |  avg   |  val  |  avg  |   val  |   avg  | val | avg |
    | ------:| ------:| -----:| -----:| ---:| ---:| ------:| ------:| -----:| -----:| ------:| ------:| ---:| ---:|
    |    4.6 |    3.5 |   0.1 |   0.1 |  49 |   2 |   52.8 |    1.1 |   0.0 |   0.9 |    211 |      4 |   5 |  18 |
    |    4.1 |    3.5 |   0.1 |   0.1 |  61 |   3 |   60.4 |    2.4 |  40.3 |   1.7 |    255 |     10 | 149 |  21 |
    ```
    """
    net  = NetworkMetrics()
    disk = DiskMetrics()

    for i in range(10000):
        time.sleep(1)

        net.update_stats()
        disk.update_stats()

        print(DUAL_METRICS_HEADER) if i % 15 == 0 else None
        row = (
            f"| {net.mb_recv_ps.val:6.1f} | {net.mb_recv_ps.avg:6.1f} "
            f"| {net.mb_sent_ps.val:5.1f} | {net.mb_sent_ps.avg:5.1f} "
            f"| {int(disk.io_util.val):3d} | {int(disk.io_util.avg):3d} "
            f"| {disk.mb_read.val:6.1f} | {disk.mb_read.avg:6.1f} "
            f"| {disk.mb_writ.val:5.1f} | {disk.mb_writ.avg:5.1f} "
            f"| {int(disk.io_read.val):6d} | {int(disk.io_read.avg):6d} "
            f"| {int(disk.io_writ.val):5d} | {int(disk.io_writ.avg):5d} "
            f"|"
        )
        print(row)

# `__all__` is left here for documentation purposes and as a
# reference to which interfaces are meant to be imported.
__all__ = [
    "usage",
]
